import types
from typing import Mapping, Type, TypeVar, Union

from marshmallow import EXCLUDE, Schema
from starlette.requests import Request

from .parser import Parser, set_parser

__all__ = ('header', 'header_schema', 'HeaderParser')

P = TypeVar('P')


def header_schema(schema: Union[Schema, Type[Schema]], cls: P,
                  unknown=EXCLUDE) -> P:
    return types.new_class('HeaderInputParams', (cls,),
                           exec_body=set_parser(HeaderParser.create(schema, unknown=unknown)))


def header(schema: Union[Schema, Type[Schema]], unknown=EXCLUDE) -> Type[Mapping]:
    return header_schema(schema, Mapping, unknown=unknown)


class HeaderParser(Parser):
    __slots__ = ()

    @property
    def location(self):
        return 'header'

    def parse(self, request: Request):
        return self.schema.load(request.headers, unknown=self.unknown)



from star_resty.method import Method
from star_resty.method.parser import RequestParser

__all__ = ('resolve_parameters', 'resolve_request_body', 'resolve_request_body_params')


def resolve_parameters(endpoint: Method):
    parameters = []
    parser = getattr(endpoint, '__parser__', None)
    if parser is None:
        return parameters

    for p in parser:
        if p.schema is not None and p.location != 'body':
            parameters.append({'in': p.location, 'schema': p.schema})

    return parameters


def resolve_request_body(endpoint: Method):
    parser = getattr(endpoint, '__parser__', None)
    if parser is None:
        return None

    content = resolve_request_body_content(parser)
    if content:
        return {'content': content}


def resolve_request_body_content(parser: RequestParser):
    content = {}
    for p in parser:
        if p.schema is not None and p.location == 'body' and p.media_type:
            content[p.media_type] = {'schema': p.schema}

    return content


def resolve_request_body_params(endpoint: Method):
    params = []
    parser = getattr(endpoint, '__parser__', None)
    if parser is None:
        return params

    for p in parser:
        if p.schema is not None and p.location == 'body' and p.media_type:
            params.append({
                'name': 'body',
                'in': 'body',
                'schema': p.schema
            })

    return params



import inspect
from typing import Union, Dict, Type

from star_resty.method import Method

__all__ = ('resolve_responses',)


def resolve_responses(endpoint: Method):
    responses = {}
    if endpoint.response_schema:
        responses[str(endpoint.status_code)] = {
            'schema': endpoint.response_schema
        }

    errors = endpoint.meta.errors or ()
    for e in errors:
        if isinstance(e, dict) and e.get('status_code'):
            responses[str(e['status_code'])] = {key: val for key, val in e.items() if key != 'status_code'}
        elif isinstance(e, Exception) and getattr(e, 'status_code', None) is not None:
            responses[str(getattr(e, 'status_code'))] = create_error_schema_by_exc(e)
        elif inspect.isclass(e) and issubclass(e, Exception) and getattr(e, 'status_code', None) is not None:
            responses[str(getattr(e, 'status_code'))] = create_error_schema_by_exc(e)

    parser = getattr(endpoint, '__parser__', None)
    if parser and '404' not in responses:
        responses['400'] = {'description': 'Bad request'}

    return responses


def create_error_schema_by_exc(e: Union[Exception, Type[Exception]]) -> Dict:
    schema = {'description': (getattr(e, 'detail', None)
                              or getattr(e, 'description', None)
                              or str(e))}
    error_schema = getattr(e, 'schema', None)
    if error_schema is not None:
        schema['schema'] = error_schema

    return schema



from starlette.routing import Route

from star_resty.method import Method
from .request import resolve_parameters, resolve_request_body, resolve_request_body_params
from .response import resolve_responses

__all__ = ('setup_route_operations',)


def setup_route_operations(route: Route, endpoint: Method, version: int = 2,
                           add_head_methods: bool = False):
    operation = setup_operation(endpoint, version=version)
    operation = {key: val for key, val in operation.items() if val is not None}
    return {method.lower(): operation for method in route.methods
            if (method != 'HEAD' or add_head_methods)}


def setup_operation(endpoint: Method, version=2):
    options = endpoint.meta
    res = {
        'tags': [options.tag],
        'description': options.description,
        'summary': options.summary,
        'produces': [endpoint.serializer.media_type],
        'parameters': resolve_parameters(endpoint),
        'responses': resolve_responses(endpoint),
    }

    if options.security is not None:
        res['security'] = options.security

    if version > 2:
        res['requestBody'] = resolve_request_body(endpoint)
    else:
        res['parameters'].extend(resolve_request_body_params(endpoint))

    if options.meta:
        res.update(options.meta)

    return res



import logging
from typing import Optional, Mapping

from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import JSONResponse

from .route import setup_routes
from .utils import resolve_schema_name

logger = logging.getLogger(__name__)

__all__ = ('setup_spec',)


def setup_spec(app: Starlette, title: str,
               version: str = '0.0.1',
               openapi_version='2.0',
               schemes=None,
               base_path='/',
               route: str = '/apidocs.json',
               add_head_methods: bool = False,
               options: Optional[Mapping] = None,
               **kwargs):
    if options is None:
        options = {}

    spec = APISpec(
        title=title,
        version=version,
        openapi_version=openapi_version,
        schemes=schemes or ['http', 'https'],
        plugins=[MarshmallowPlugin(schema_name_resolver=resolve_schema_name)],
        **{'swagger': openapi_version, 'basePath': base_path, **options, **kwargs}
    )
    api_spec = None

    @app.route(route, include_in_schema=False)
    def generate_api_docs(_: Request):
        nonlocal api_spec
        nonlocal spec
        if api_spec is None:
            logger.info('initialize open api schema')
            setup_routes(app.routes, spec, version=get_open_api_version(openapi_version)
                         , add_head_methods=add_head_methods)
            api_spec = spec.to_dict()

        return JSONResponse(api_spec)


def get_open_api_version(version: str) -> int:
    v = version.split('.', maxsplit=1)[0]
    try:
        return int(v)
    except (ValueError, TypeError):
        raise ValueError(f'Invalid open api version: {version}')



import abc
import inspect
from typing import Dict, Optional, Type, Union

from marshmallow import EXCLUDE, Schema
from starlette.requests import Request
from functools import lru_cache

__all__ = ('Parser', 'set_parser')


class Parser(abc.ABC):
    __slots__ = ('schema', 'unknown')

    @classmethod
    def create(cls, schema: Union[Schema, Type[Schema]], unknown: str = EXCLUDE):
        return cls(cls._convert_schema(schema), unknown)

    @staticmethod
    @lru_cache(maxsize=1024)
    def _convert_schema(schema: Union[Schema, Type[Schema]]) -> Schema:
        if inspect.isclass(schema):
            if not issubclass(schema, Schema):
                raise TypeError(f'Invalid schema type: {schema}')

            return schema()
        elif not isinstance(schema, Schema):
            raise TypeError(f'Invalid schema type: {type(schema)}')

        return schema

    def __init__(self, schema: Schema, unknown=EXCLUDE):
        self.schema = schema
        self.unknown = unknown

    @abc.abstractmethod
    def parse(self, request: Request):
        pass

    @property
    def location(self) -> Optional[str]:
        return None

    @property
    def media_type(self) -> Optional[str]:
        return None


def set_parser(parser: Parser):
    def handler(ns: Dict):
        ns['parser'] = parser
        return ns

    return handler



import inspect
import itertools
from dataclasses import is_dataclass, fields
from functools import partial
from typing import Dict, Tuple, Sequence, Generator, Mapping, Generic, TypeVar, Type, Callable, Optional, Any, Union

from starlette.requests import Request

from star_resty.payload.parser import Parser

__all__ = ('RequestParser', 'create_parser')

D = TypeVar('D')


class RequestParser:
    __slots__ = ('_parsers', '_async_parsers')

    def __init__(self, parsers: Sequence[Tuple[str, Union[Parser, 'RequestParser']]] = (),
                 async_parsers: Sequence[Tuple[str, Union[Parser, 'RequestParser']]] = ()):
        self._parsers = parsers
        self._async_parsers = async_parsers

    def __nonzero__(self):
        return bool(self._parsers or self._async_parsers)

    def __iter__(self) -> Generator[Parser, None, None]:
        parsers = itertools.chain(self._parsers, self._async_parsers)
        for (_, p) in parsers:
            if isinstance(p, RequestParser):
                yield from p
            else:
                yield p

    async def parse(self, request: Request) -> Dict:
        params = {}
        for (key, p) in self._parsers:
            params[key] = p.parse(request)

        for (key, p) in self._async_parsers:
            params[key] = await p.parse(request)

        return params


class DataClassParser(RequestParser, Generic[D]):
    __slots__ = ('_data_cls',)

    def __init__(self, data_cls: Type[D], parsers: Sequence[Tuple[str, Parser]] = (),
                 async_parsers: Sequence[Tuple[str, Parser]] = ()):
        super().__init__(parsers=parsers, async_parsers=async_parsers)
        self._data_cls = data_cls

    async def parse(self, request: Request) -> D:
        params = await super().parse(request)
        return self._data_cls(**params)


def create_parser(func) -> RequestParser:
    if func is None:
        return RequestParser()

    return create_parser_from_data(func.__annotations__)


def create_parser_from_data(data: Mapping) -> RequestParser:
    parsers = []
    async_parsers = []
    for key, value in data.items():
        if is_dataclass(value):
            data = {field.name: field.type for field in fields(value)}
            factory = partial(DataClassParser, value)
            parser = create_parser_for_dc(data, factory=factory)
        else:
            parser = getattr(value, 'parser', None)

        if parser is None or not isinstance(parser, (Parser, RequestParser)):
            continue

        if inspect.iscoroutinefunction(parser.parse):
            async_parsers.append((key, parser))
        else:
            parsers.append((key, parser))

    return RequestParser(parsers=parsers, async_parsers=async_parsers)


def create_parser_for_dc(data: Mapping, factory: Callable) -> RequestParser:
    parsers = []
    async_parsers = []
    for key, value in data.items():
        if is_dataclass(value):
            data = {field.name: field.type for field in fields(value)}
            factory = partial(DataClassParser, value)
            parser = create_parser_for_dc(data, factory=factory)
        else:
            parser = getattr(value, 'parser', None)

        if parser is None:
            parser = get_parser_from_args(value)

        if parser is None or not isinstance(parser, (Parser, RequestParser)):
            continue

        if inspect.iscoroutinefunction(parser.parse):
            async_parsers.append((key, parser))
        else:
            parsers.append((key, parser))

    return factory(parsers=parsers, async_parsers=async_parsers)


def get_parser_from_args(value: Any) -> Optional[Parser]:
    args = getattr(value, '__args__', None) or ()
    for a in args:
        parser = getattr(a, 'parser', None)
        if parser is not None and isinstance(parser, Parser):
            return parser



import inspect
import logging
from typing import Sequence, Callable, Any

from marshmallow import Schema
from marshmallow.exceptions import MarshmallowError
from starlette.responses import Response

from star_resty.exceptions import DumpError

__all__ = ('create_render', 'Render')

logger = logging.getLogger(__name__)


def create_render(method) -> 'Render':
    renders = []
    response_schema = getattr(method, 'response_schema', None)
    if response_schema is None:
        response_schema = getattr(method, 'Response', None)

    if response_schema is not None:
        if inspect.isclass(response_schema):
            response_schema = response_schema()
        renders.append(dump_content(response_schema, None))

    serializer = getattr(method, 'serializer', None)
    if serializer is not None:
        renders.append(render_bytes(serializer, method))

    return Render(renders)


class Render:
    __slots__ = ('_renders',)

    def __init__(self, renders: Sequence):
        self._renders = renders

    def __call__(self, content: Any, method):
        for r in self._renders:
            content = r(content, method)

        return content


def render_bytes(serializer, method):
    def render(content, method):
        return Response(serializer.render(content),
                        media_type=serializer.media_type,
                        status_code=method.status_code)

    return render


def dump_content(response_schema: Schema, _) -> Callable:
    def dump(content, _):
        try:
            return response_schema.dump(content)
        except MarshmallowError as e:
            logger.error('Dump error: %s', e)
            raise DumpError(e) from e
        except (ValueError, TypeError) as e:
            logger.error('Dump error: %s', e)
            raise DumpError(e) from e

    return dump



import abc

from .parser import create_parser
from .render import create_render

__all__ = ('MethodMeta',)


class MethodMeta(abc.ABCMeta):

    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super().__new__(mcs, name, bases, attrs, **kwargs)

        func = getattr(cls, 'execute', None)
        if func is None:
            raise TypeError(f'Invalid method class={name}')

        cls.__parser__ = create_parser(func)
        cls.__render__ = create_render(cls)
        return cls



import abc
from functools import wraps
from typing import ClassVar, Type, Union, Optional, Callable, Any

from marshmallow import Schema
from starlette.requests import Request
from starlette.responses import Response

from star_resty.operation import Operation
from star_resty.serializers import JsonSerializer, Serializer
from .meta import MethodMeta

__all__ = ('Method', 'endpoint')


class Method(abc.ABC, metaclass=MethodMeta):
    __slots__ = ('request',)
    __parser__: ClassVar[Callable[[Request], Any]]
    __render__: ClassVar[Callable[[Any], Response]]

    meta: ClassVar[Operation] = Operation(tag='default')
    serializer: ClassVar[Serializer] = JsonSerializer
    response_schema: ClassVar[Union[Schema, Type[Schema], None]] = None
    Response: ClassVar[Optional[Type[Schema]]] = None
    status_code: int = 200

    def __init__(self, request: Request):
        self.request = request

    @abc.abstractmethod
    async def execute(self, *args, **kwargs):
        pass

    async def dispatch(self) -> Response:
        params = await self.__parser__.parse(self.request)
        content = await self.execute(**params)
        return self.__render__(content, self)

    @classmethod
    def as_endpoint(cls):
        return endpoint(cls)


def endpoint(cls: Type[Method]):
    @wraps(cls)
    async def wrapper(request: Request) -> Response:
        method = cls(request)
        return await method.dispatch()

    wrapper.__endpoint__ = cls
    return wrapper



from typing import Sequence

from apispec import APISpec
from starlette.routing import BaseRoute, Mount, Route

from .operation import setup_route_operations
from .utils import convert_path

__all__ = ('setup_routes',)


def setup_routes(routes: Sequence[BaseRoute],
                 spec: APISpec, version: int = 2,
                 add_head_methods: bool = False,
                 path: str = ''):
    for route in routes:
        if isinstance(route, Mount):
            setup_routes(route.routes, spec, version=version, add_head_methods=add_head_methods,
                         path=f'{path}{route.path}')
            continue
        elif isinstance(route, Route):
            if not route.include_in_schema:
                continue

            endpoint = getattr(route.endpoint, '__endpoint__', None)
            if endpoint is None:
                continue

            operations = setup_route_operations(route, endpoint, version=version,
                                                add_head_methods=add_head_methods)
            route_path = f'{path}{route.path}'
            spec.path(convert_path(route_path), operations=operations)



from dataclasses import dataclass
from typing import Optional, Sequence, Any, Mapping

__all__ = ('Operation',)


@dataclass(frozen=True)
class Operation:
    tag: str = 'default'
    description: Optional[str] = None
    summary: Optional[str] = None
    errors: Sequence[Any] = ()
    security: Optional[Sequence] = None
    meta: Optional[Mapping] = None

    @classmethod
    def create(cls,
               tag: str = 'default',
               description: Optional[str] = None,
               summary: Optional[str] = None,
               errors: Sequence[Any] = (),
               security: Optional[Sequence] = None,
               **kwargs) -> 'Operation':
        return cls(tag=tag, description=description,
                   summary=summary, errors=errors,
                   security=security, meta=kwargs)



import ujson

__all__ = ('JsonSerializer',)


class JsonSerializer:
    media_type = 'application/json'

    @staticmethod
    def render(content) -> bytes:
        return ujson.dumps(content, ensure_ascii=False).encode('utf-8')



from typing import Dict, Optional

from marshmallow.exceptions import ValidationError

__all__ = ('StarRestError', 'DumpError', 'DecodeError')


class StarRestError(Exception):
    pass


class DumpError(StarRestError):

    def __init__(self, exc: Optional[Exception] = None, *args):
        super().__init__(*args)
        self.orig_exc = exc

    def normalized_messages(self) -> Dict:
        if self.orig_exc is not None and isinstance(self.orig_exc, ValidationError):
            return self.orig_exc.normalized_messages()

        return {'_schema': 'Error dump content'}


class DecodeError(StarRestError):

    def normalized_messages(self) -> Dict:
        return {'_body': str(self)}



import types
from typing import Mapping, Type, TypeVar, Union

import ujson
from marshmallow import EXCLUDE, Schema
from starlette.requests import Request

from star_resty.exceptions import DecodeError
from .parser import Parser, set_parser

__all__ = ('json_schema', 'json_payload', 'JsonParser')

P = TypeVar('P')


def json_schema(schema: Union[Schema, Type[Schema]], cls: P,
                unknown: str = EXCLUDE) -> P:
    return types.new_class('JsonInputParams', (cls,),
                           exec_body=set_parser(JsonParser.create(schema, unknown=unknown)))


def json_payload(schema: Union[Schema, Type[Schema]], unknown=EXCLUDE) -> Type[Mapping]:
    return json_schema(schema, Mapping, unknown=unknown)


class JsonParser(Parser):
    __slots__ = ()

    @property
    def location(self):
        return 'body'

    @property
    def media_type(self):
        return 'application/json'

    async def parse(self, request: Request):
        body = await request.body()
        if body is None:
            data = {}
        else:
            try:
                data = ujson.loads(body)
            except (TypeError, ValueError) as e:
                raise DecodeError('Invalid json body') from e

        return self.schema.load(data, unknown=self.unknown)



import operator
from typing import TypeVar, Type, Optional, Generic

from star_resty import Method

__all__ = ('attr',)

T = TypeVar('T')


class InjectAttr(Generic[T]):
    __slots__ = ('_func',)

    def __init__(self, name=None):
        if name:
            self._func = operator.attrgetter(name)
        else:
            self._func = None

    def __set_name__(self, owner, name):
        if self._func is None:
            self._func = operator.attrgetter(name)

    def __get__(self, instance: Optional[Method], owner: Type[Method]) -> T:
        if instance is None:
            return self

        return self._func(instance.request.app)


def attr(_: Optional[Type[T]] = None, *, name: Optional[str] = None) -> T:
    return InjectAttr[T](name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import setup


def get_long_description():
    with open('README.md', encoding='utf8') as f:
        return f.read()


def get_packages(package):
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, '__init__.py'))
    ]


setup(
    name='star_resty',
    python_requires='>=3.7',
    install_requires=[
        'ujson',
        'typing_extensions',
        'marshmallow>=3.0.0rc8,<4',
        'starlette<1',
        'apispec<4',
    ],
    version='0.0.14',
    url='https://github.com/slv0/start_resty',
    license='BSD',
    description='The web framework',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    author='Slava Cheremushkin',
    author_email='slv0.chr@gmail.com',
    packages=get_packages('star_resty'),
    package_data={'star_resty': ['py.typed']},
    data_files=[('', ['LICENSE'])],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python :: 3.7',
    ],
    zip_safe=False,
)



import inspect
import types
from functools import lru_cache
from typing import Mapping, Type, TypeVar, Union, Callable, Sequence, List, Tuple, Iterator

from marshmallow import EXCLUDE, Schema, fields
from starlette.requests import Request

from .parser import Parser, set_parser

__all__ = ('query', 'query_schema', 'QueryParser')

Q = TypeVar('Q')


def query_schema(schema: Union[Schema, Type[Schema]], cls: Q,
                 unknown=EXCLUDE) -> Q:
    return types.new_class('QueryInputParams', (cls,),
                           exec_body=set_parser(QueryParser.create(schema, unknown=unknown)))


def query(schema: Union[Schema, Type[Schema]], unknown=EXCLUDE) -> Type[Mapping]:
    return query_schema(schema, Mapping, unknown=unknown)


class QueryParser(Parser):
    __slots__ = ('fields',)

    @classmethod
    def create(cls, schema: Union[Schema, Type[Schema]], unknown: str = EXCLUDE):
        schema, query_fields = get_query_fields(schema)
        return cls(schema, query_fields, unknown)

    def __init__(self, schema: Schema, query_fields: Mapping, unknown=EXCLUDE):
        super().__init__(schema, unknown=unknown)
        self.fields = query_fields

    @property
    def location(self):
        return 'query'

    def parse(self, request: Request):
        query_params = request.query_params
        getlist = request.query_params.getlist
        query_fields = self.fields
        data = ((key, query_fields[key](getlist(key)))
                for key in query_params.keys() if key in query_fields)
        data = {key: val for (key, val) in data if val is not None}
        return self.schema.load(data, many=False, unknown=self.unknown)


@lru_cache(typed=False, maxsize=1024)
def get_query_fields(schema: Union[Schema, Type[Schema]]) -> Tuple[Schema, Mapping[str, Callable]]:
    if inspect.isclass(schema):
        schema = schema()

    query_fields = dict(iter_query_fields(schema))
    return schema, query_fields


def iter_query_fields(schema: Schema) -> Iterator[Tuple[str, Callable]]:
    for key, field in schema.fields.items():
        field: fields.Field
        if field.dump_only or isinstance(field, fields.Constant):
            continue

        if field.attribute:
            name = field.attribute
        else:
            name = key

        if isinstance(field, fields.List):
            func = get_list_value
        else:
            func = get_value

        yield name, func
        if field.data_key:
            yield field.data_key, func


def get_list_value(values: Sequence) -> List:
    return [v for v in values if v]


def get_value(values: Sequence):
    return next((v for v in values if v), None)



import inspect
import re
from typing import Any

__all__ = ('resolve_schema_name', 'convert_path')


def resolve_schema_name(schema: Any) -> str:
    if inspect.isclass(schema):
        cls = schema
    else:
        cls = schema.__class__

    name = f'{cls.__module__}.{cls.__qualname__}'
    return name


def convert_path(path: str) -> str:
    return re.sub(r'{([^:]+).*}', r'{\1}', path)
